ORGANISM <- "Oreochromis niloticus"

### List of assemblies by date.
ASSEMBLIES <- list(
    ## 5678 sequences.
    list(assembly="Orenil1.1",
         assembly_level="Chromosome",
         date="2012/02/08",
         ## GCA_000188235.2 and GCF_000188235.2 are identical.
         assembly_accession="GCA_000188235.2",  # oreNil2
         circ_seqs="MT"),

    ## 2460 sequences.
    list(assembly="O_niloticus_UMD_NMBU",
         assembly_level="Chromosome",
         date="2018/06/29",
         ## GCA_001858045.3 and GCF_001858045.2 are identical.
         assembly_accession="GCF_001858045.2",
         circ_seqs="MT")
)

