ORGANISM <- "Gadus morhua"

### List of assemblies by date.
ASSEMBLIES <- list(
    ## 427427 sequences.
    list(assembly="GadMor_May2010",
         assembly_level="Scaffold",
         date="2011/08/26",
         assembly_accession="GCA_000231765.1",  # gadMor1
         circ_seqs=character(0)),

    ## 227 sequences.
    list(assembly="gadMor3.0",
         assembly_level="Chromosome",
         date="2019/07/16",
         ## GCA_902167405.1 and GCF_902167405.1 are identical.
         assembly_accession="GCF_902167405.1",
         circ_seqs="MT")
)

