ORGANISM <- "Rattus norvegicus"

### List of assemblies by date.
ASSEMBLIES <- list(
    list(assembly="RGSC_v3.4",
         date="2004/12/13",
         assembly_accession="GCF_000001895.3",  # rn4
         circ_seqs="MT"),

    list(assembly="Rnor_5.0",
         date="2012/03/16",
         assembly_accession="GCA_000001895.3",  # rn5
         circ_seqs="MT"),

    list(assembly="Rnor_6.0",
         date="2014/07/01",
         assembly_accession="GCA_000001895.4",  # rn6
         circ_seqs="MT"),

    list(assembly="mRatBN7.2",
         date="2020/11/10",
         assembly_accession="GCF_015227675.2",  # rn7
         circ_seqs="MT")
)

