ORGANISM <- "Gasterosteus aculeatus"

### List of assemblies by date.
ASSEMBLIES <- list(
    ## 16957 sequences.
    list(assembly="ASM18067v1",
         assembly_level="Contig",
         date="2006/02/01",
         assembly_accession="GCA_000180675.1",  # gasAcu1
         circ_seqs=character(0)),

    ## 2938 sequences.
    list(assembly="GAculeatus_UGA_version5",
         assembly_level="Chromosome",
         date="2021/02/23",
         ## GCA_016920845.1 and GCF_016920845.1 are identical.
         assembly_accession="GCF_016920845.1",
         circ_seqs="MT")
)

