ORGANISM <- "Heterocephalus glaber"

### List of assemblies by date.
ASSEMBLIES <- list(
    ## 39267 sequences.
    list(assembly="HetGla_1.0",
         assembly_level="Scaffold",
         date="2011/10/14",
         assembly_accession="GCA_000230445.1",  # hetGla1
         circ_seqs="MT"),

    ## 4229 sequences.
    list(assembly="HetGla_female_1.0",
         assembly_level="Scaffold",
         date="2012/02/17",
         ## GCA_000247695.1 and GCF_000247695.1 are identical.
         assembly_accession="GCF_000247695.1",  # hetGla2
         circ_seqs="MT"),

    ## 32 sequences.
    list(assembly="Naked mole-rat maternal",
         assembly_level="Chromosome",
         date="2022/07/01",
         assembly_accession="GCA_944319715.1",
         circ_seqs="MT"),

    ## 32 sequences.
    list(assembly="Naked mole-rat paternal",
         assembly_level="Chromosome",
         date="2022/07/01",
         assembly_accession="GCA_944319725.1",
         circ_seqs="MT")
)

