\name{GenomeInfoDb internals}

\alias{DEFAULT_CIRC_SEQS}
\alias{list_ftp_dir}

\title{GenomeInfoDb internals}

\description{
  Symbols defined in the \pkg{GenomeInfoDb} package that are not
  intended to be used directly.
}

\keyword{internal}
